import secrets
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from werkflow.connections.env import Env


class AESGCMFernet:

    def __init__(self, env: Env) -> None:
        self.secret = env.MERCURY_SYNC_AUTH_SECRET

    def encrypt(self, data: bytes) -> bytes:
        key = secrets.token_bytes(32)
        nonce = secrets.token_bytes(12)
        return key + nonce + AESGCM(key).encrypt(nonce, data, b"")
    
    def decrypt(self, data: bytes) -> bytes:
        key = data[:32]
        nonce = data[32:44]
        return AESGCM(key).decrypt(nonce, data[44:], b"")
    



import click
import asyncio
import functools
from typing import Dict, Type, Optional, Any, Callable, Union
from werkflow.prompt.types.base.base_prompt import BasePrompt
from werkflow.prompt.types.base.prompt_type import PromptType
from .input_prompt_validator import InputPromptValidator


class InputPrompt(BasePrompt):

    def __init__(
        self, 
        message: str, 
        skip: bool = False, 
        result_key: str = None, 
        data_type: Type = str,
        default: Optional[Any] = None, 
        condition: Optional[Callable[[Dict[str, Any]], bool]] = None, 
        confirmation_message: Optional[Union[str, Callable[..., str]]] = None, 
    ) -> None:
        super().__init__(
            message, 
            skip, 
            result_key, 
            default, 
            condition, 
            confirmation_message,
        )

        validated_prompt = InputPromptValidator(
            prompt_data_type=data_type,
            prompt_type=PromptType.INPUT
        )

        self.data_type = validated_prompt.prompt_data_type
        self.prompt_type = validated_prompt.prompt_type
        

    async def ask(self):
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        return await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                click.prompt,
                f'{self.prompt_frame} {self.message}',
                type=self.data_type,
                default=self.default,
                prompt_suffix=''
            )
        )


import click
import functools
from werkflow.prompt.types.base.base_prompt import BasePrompt
from werkflow.prompt.types.base.prompt_type import PromptType
from typing import Callable, Union, Optional, Any, Dict
from .key_value_prompt_validator import KeyValuePromptValidator


class KeyValuePrompt(BasePrompt):

    def __init__(
        self,
        key_prompt: BasePrompt=None,
        value_prompt: BasePrompt=None,
        result_key: str=None,
        skip: bool=False,
        condition: Optional[Callable[[Dict[str, Any]], bool]]=None

    ) -> None:
        
        super().__init__(
            None,
            skip,
            result_key,
            None,
            condition,
            None
        )

        validated_prompt = KeyValuePromptValidator(
            key_prompt=key_prompt,
            value_prompt=value_prompt,
            prompt_result_key=result_key,
            prompt_type=PromptType.KEY_VALUE,
            prompt_condition=condition
        )

        self.key_prompt: BasePrompt = validated_prompt.key_prompt
        self.value_prompt: BasePrompt = validated_prompt.value_prompt
        self.prompt_type = validated_prompt.prompt_type

    async def ask(self):    
        key_response = await self.key_prompt.ask()

        if key_response == self.key_prompt.default:
            return {
                'key': key_response,
                'value': None
            }

        await self.key_prompt.confirm(key_response)

        value_response = await self.value_prompt.ask()
        await self.value_prompt.confirm(value_response)

        return {
            'key': key_response,
            'value': value_response
        }

    async def confirm(self, _: Any):
        pass

    def close(self):
        self.key_prompt.close()
        self.value_prompt.close()


import asyncio
import subprocess
import functools
import pathlib
import os
from concurrent.futures import ThreadPoolExecutor
from werkflow.modules.system import System
from typing import Dict, Union, TextIO, BinaryIO, Optional


class ZSHShell:

    def __init__(self) -> None:
        self.last_command: str = None
        self.env = os.environ.copy()
        system = System()
        self._executor = ThreadPoolExecutor(
            max_workers=system.configuration.cores.physical
        )

        self._loop: asyncio.AbstractEventLoop = None
        self.envfile_name = '.zshrc'
        self.envfile_path = os.path.join(
            system.users.home,
            self.envfile_name
        )

        self.silent = False

    def path_is_file(
        self,
        path: str
    ) -> bool:
        return len(pathlib.Path(path).suffix) > 0
    
    def path_is_directory(
        self,
        path: str
    ) -> bool:
        return len(pathlib.Path(path).suffix) == 0
    
    async def get_envar(
        self,
        envar_name: str
    ) -> Union[str, None]:
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        return await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.getenv,
                envar_name
            )
        )
    
    async def set_envar(
        self,
        envar_name: str,
        envar_value: str
    ):
        
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.putenv,
                envar_name,
                envar_value
            )
        )


    async def check_if_is_file(
        self,
        filepath: str
    ) -> bool:
        if self._loop is None:
            self._loop = asyncio.get_running_loop()
            
        return await self._loop.run_in_executor(
            self._executor,
            functools.partial(   
                os.path.isfile,
                filepath
            )
        )
    
    async def check_if_is_directory(
        self,
        filepath: str
    ) -> bool:
        if self._loop is None:
            self._loop = asyncio.get_running_loop()
            
        return await self._loop.run_in_executor(
            self._executor,
            functools.partial(   
                os.path.isdir,
                filepath
            )
        )

    async def to_absolute_path(
        self, 
        filepath: str
    ) -> str:

        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        system = System()
        if "~/" == filepath[:2]:
            filepath = os.path.join(
                system.users.home,
                filepath[2:]
            )

        path = pathlib.Path(filepath)
        resolved_path = await self._loop.run_in_executor(
            self._executor,
            path.resolve
        )

        absolute_path = str(resolved_path)

        if filepath[-1] == "/":
            absolute_path = f'{absolute_path}/'
        
        return absolute_path
    
    async def to_relative_path(
        self,
        base_path: str,
        absolute_path: str
    ) -> str:
        
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        relative_path = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.relpath,
                absolute_path,
                base_path
            )
        )

        return relative_path.replace('../', '')

    async def get_current_directory(self) -> str:

        if self._loop is None:
            self._loop = asyncio.get_running_loop()
            
        return await self._loop.run_in_executor(
            self._executor,
            os.getcwd
        )

    async def change_directory(
        self,
        directory_path: str
    ) -> None:

        absolute_path = await self.to_absolute_path(directory_path)

        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.chdir,
                absolute_path
            )
        )

    async def assume_admin(
        self,
        password: str=None
    ) -> subprocess.CompletedProcess[str]:
        
        output_pipe = subprocess.PIPE if self.silent else None

        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        result = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                subprocess.run,
                [*f"sudo -S -v".split()],
                stdout=output_pipe, 
                stderr=output_pipe,
                text=True, 
                input=password,
                env={
                    **self.env
                }
            )
        )

        return result

    async def run(self, 
            command: str,
            command_input: str=None,
            path: Optional[str]=None,
            env: Dict[str, str]={},
            as_shell: bool=False
        ) -> subprocess.CompletedProcess[str]:
        
        output_pipe = subprocess.PIPE if self.silent else None
        
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        process_command = command
        if as_shell is False:
            process_command = [*command.split()]

        self.last_command = command
        result = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                subprocess.run,
                process_command,
                cwd=path,
                stdout=output_pipe, 
                stderr=output_pipe,
                text=True, 
                input=command_input,
                shell=as_shell,
                env={
                    **self.env,
                    **{
                        key: str(value) for key, value in env.items()
                    }
                }
            )
        )

        return result
    
    async def eval(
        self,
        command: str,
        env: Dict[str, str]={}
    ) -> None:
        await self.run(
            f'eval "$({command})"',
            env=env,
            as_shell=True
        )

    async def copy(
        self,
        source_directory: str,
        destination_directory: str,
        recursive: bool=False
    ) -> None:
        
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        if recursive:
            await self.run(
                f'cp -r {source_directory} {destination_directory}'
            )

        else:
            await self.run(
                f'cp {source_directory} {destination_directory}'
            )
    
    
    async def move(
        self,
        source_directory: str,
        destination_directory: str
    ) -> None:
        
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        await self.run(
            f'mv {source_directory} {destination_directory}',
            as_shell=True
        )
    
    async def append_to_envfile(
        self,
        variable_name: str,
        variable_value: str
    ) -> None:
        
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        environmental_variable = f'export {variable_name}={variable_value}'
        envfile = await self.open_or_create_file(self.envfile_path)

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                envfile.write,
                environmental_variable
            )
        )

        await self._loop.run_in_executor(
            self._executor,
            envfile.close
        )

    async def check_path_exists(
        self,
        filepath: str
    ) -> bool:
        
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        
        absolute_path = await self.to_absolute_path(filepath)

        return await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.exists,
                absolute_path
            )
        )
    
    async def touch(
        self,
        filepath: str,
        is_bytes: bool=False,
        overwrite: bool=False
    ) -> None:
        
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        absolute_path = await self.to_absolute_path(filepath)

        created_file = await self.create_file(
            absolute_path,
            is_bytes=is_bytes,
            overwrite=overwrite
        )
        
        if created_file:
            await self._loop.run_in_executor(
                self._executor,
                created_file.close
            )

    async def create_file(
        self,
        filepath: str,
        is_bytes: bool=False,
        overwrite: bool=False
    ) -> Union[TextIO, BinaryIO]:
        
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        absolute_path = await self.to_absolute_path(filepath)
        file_exists = await self.check_path_exists(absolute_path)

        write_mode = 'w+'
        if is_bytes:
            write_mode = 'wb+'

        if file_exists is False or overwrite is True:
            return await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    absolute_path,
                    write_mode
                )
            )
        
    async def open_or_create_file(
        self,
        filepath: str,
        is_bytes: bool=False,
        overwrite: bool=False
    ) -> Union[TextIO, BinaryIO]:
        
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        absolute_path = await self.to_absolute_path(filepath)

        write_file = await self.create_file(
            absolute_path,
            is_bytes=is_bytes,
            overwrite=overwrite
        )

        if write_file is None:
            write_file = await self.open_existing_file(
                absolute_path,
                is_bytes=is_bytes,
            )

        return write_file
        
    async def open_existing_file(
        self,
        filepath: str,
        is_bytes: bool=False,
        write_mode: str='a+'
    ) -> Union[TextIO, BinaryIO]:
        
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        if is_bytes and write_mode == 'a+':
            write_mode = 'ab+'

        elif is_bytes and write_mode == 'w+':
            write_mode = 'wb+'

        elif is_bytes and write_mode == 'r':
            write_mode = 'rb'

        absolute_path = await self.to_absolute_path(filepath)

        return await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                open,
                absolute_path,
                write_mode
            )
        )
    
    async def read_file(      
        self,
        filepath: str,
        is_bytes: bool=False
    ):

        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        absolute_path = await self.to_absolute_path(filepath)

        file_data = await self.open_existing_file(
            absolute_path,
            write_mode='r',
            is_bytes=is_bytes
        )

        data = await self._loop.run_in_executor(
            self._executor,
            file_data.read
        )

        await self._loop.run_in_executor(
            self._executor,
            file_data.close
        )

        return data

    async def pipe_to_file(
        self,
        filepath: str,
        data: Union[str, bytes],
        overwrite: bool=False
    ) -> None:
        
        if self._loop is None:
            self._loop = asyncio.get_running_loop()
        
        is_bytes = False
        if isinstance(data, bytes):
            is_bytes = True

        absolute_path = await self.to_absolute_path(filepath)

        write_file = await self.open_or_create_file(
            absolute_path,
            is_bytes=is_bytes,
            overwrite=overwrite
        )

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                write_file.write,
                data
            )
        )

        await self._loop.run_in_executor(
            self._executor,
            write_file.close
        )

    async def create_directory(
        self,
        directory_path: str
    ) -> None:
        
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        absolute_path = await self.to_absolute_path(directory_path)

        path_exists = await self.check_path_exists(absolute_path)

        if path_exists is False:
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    os.makedirs,
                    absolute_path
                )
            )

    async def source(self, filepath: str=None) -> None:

        if filepath is None:
            filepath = self.envfile_path

        absolute_path = await self.to_absolute_path(filepath)
        await self.run(
            f'source {absolute_path}',
            as_shell=True
        )

    async def rm(
        self,
        path: str,
        pattern: Optional[str]=None,
        recurse: bool=False
    ) -> None:
        
        absolute_path = await self.to_absolute_path(path)

        path_exists = await self.check_path_exists(absolute_path)

        if pattern is None:
            pattern = '*'

        if recurse and path_exists:
            recurse_path = pathlib.Path(absolute_path)

            matching = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    recurse_path.rglob,
                    pattern
                )
            )

            matching_items = [
                str(matching_path) for matching_path in matching
            ]

            if len(matching_items) > 0:
                await asyncio.gather(
                    self._loop.run_in_executor(
                        self._executor,
                        functools.partial(
                            os.remove,
                            matching_path
                        )
                    ) for matching_path in matching_items
                )

        elif path_exists:
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    os.remove,
                    path
                )
            )


    async def close(self) -> None:
        self._executor.shutdown()

    def abort(self):
        self._executor.shutdown()

        


import os
import asyncio
import datetime
from typing import Dict
from pathlib import Path
from aiologger.levels import LogLevel
from aiologger.formatters.base import Formatter
from werkflow.tools.filesystem import open
from .handlers.async_file_handler import AsyncTimedRotatingFileHandler, RolloverInterval
from .logger_types import LoggerTypes
from .async_logger import AsyncLogger

class AsyncFilesystemLogger:

    def __init__(
        self, 
        logger_name: str=None,
        logger_type: LoggerTypes=LoggerTypes.FILESYSTEM,
        logfiles_directory: str=None,
        log_level: LogLevel=LogLevel.NOTSET, 
        logger_enabled: bool = True,
        rotation_interval_type: RolloverInterval=RolloverInterval.DAYS,
        rotation_interval: int=1,
        backup_count: int=1,
        rotation_time: datetime.time=None
    ) -> None:
        self.logger_name = logger_name
        self.logger_type = logger_type
        self.log_level = log_level
        self.logger_enabled = logger_enabled
        self.rotation_interval_type = rotation_interval_type
        self.rotation_interval = rotation_interval
        self.backups = backup_count
        self.rotation_time = rotation_time
        self.logfiles_directory: str = logfiles_directory

        self.files: Dict[str, AsyncLogger] = {}
        self.filepaths: Dict[str, str] = {}

    def __getitem__(self, logger_name: str):

        file_logger = self.files.get(logger_name)

        if file_logger is None:
            file_logger = self._create_file_logger(
                logger_name,
                os.path.join(
                    self.logfiles_directory,
                    f'{logger_name}.log'
                ) 
            )

            self.files[logger_name] = file_logger

        return file_logger

    def _create_file_logger(self, logger_name: str, filepath: str) -> AsyncLogger:
        async_logger = AsyncLogger(
            logger_name=logger_name,
            logger_type=self.logger_type,
            log_level=self.log_level,
            logger_enabled=self.logger_enabled
        )

        async_file_handler = AsyncTimedRotatingFileHandler(
            filepath,
            when=self.rotation_interval_type,
            interval=self.rotation_interval,
            backup_count=self.backups,
            at_time=self.rotation_time
        )

        async_file_handler.formatter = Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(module)s:%(funcName)s:%(lineno)d - %(message)s',
            '%Y-%m-%dT%H:%M:%S.%Z'
        )

        async_logger.add_handler(async_file_handler)

        return async_logger

    async def create_logfile(self, log_filename: str):

        filepath = os.path.join(self.logfiles_directory, log_filename)

        loop = asyncio.get_event_loop()
        path_exists = await loop.run_in_executor(
            None,
            os.path.exists,
            filepath
        )
        
        if path_exists is False:
            logfile = await open(filepath, 'w')
            await logfile.close()

            self.update_files(filepath)

        else:
            self.update_files(filepath)


    def update_files(self, filepath: str):
        logger_name = Path(filepath).stem

        if self.files.get(logger_name) is None:
            self.files[logger_name] = self._create_file_logger(logger_name, filepath)
            self.filepaths[logger_name] = filepath




from typing import Dict
from typing import Generic, TypeVar
from werkflow.logging.config import LoggingConfig
from werkflow.logging.spinner.progress_text import ProgressText
from .async_filesystem_logger import AsyncFilesystemLogger
from .async_logger import AsyncLogger
from .async_spinner import AsyncSpinner
from .sync_filesystem_logger import SyncFilesystemLogger
from .sync_logger import SyncLogger
from .logger_types import LoggerTypes
from .logger_types_map import LoggerTypesMap


A = TypeVar('A')
S = TypeVar('S')


class Logger(Generic[A, S]):

    def __init__(self, config: LoggingConfig) -> None:
        self.logger_types = LoggerTypesMap()
        self.log_level = config.log_level
        self.logger_enabled = config.logger_enabled

        if config.logger_type == LoggerTypes.SPINNER:
            self.aio: A = self.logger_types.async_loggers.get(config.logger_type, AsyncSpinner)(**config.spinner)
            self.sync: S = None

        elif config.logger_type == LoggerTypes.FILESYSTEM or config.logger_type == LoggerTypes.DISTRIBUTED_FILESYSTEM:
            self.aio: A = self.logger_types.async_loggers.get(config.logger_type, AsyncFilesystemLogger)(**config.filesystem_logger)
            self.sync: S = self.logger_types.sync_loggers.get(config.logger_type, SyncFilesystemLogger)(**config.filesystem_logger)

        else:
            self.aio: A = self.logger_types.async_loggers.get(config.logger_type, AsyncLogger)(**config.cli_logger)
            self.sync: S = self.logger_types.sync_loggers.get(config.logger_type, SyncLogger)(**config.cli_logger)

    def set_patterns(self, pattern: str, datefmt_pattern: str=None):

        if isinstance(self.aio, AsyncLogger) and isinstance(self.sync, SyncLogger):
            self.aio.initialize(
                pattern,
                datefmt_pattern=datefmt_pattern
            )

            self.sync.initialize(
                pattern,
                datefmt_pattern=datefmt_pattern
            )

    def create_filelogger(self, filepath: str):
        if isinstance(self.aio, AsyncFilesystemLogger) and isinstance(self.sync, SyncFilesystemLogger):
            self.aio.update_files(filepath)
            self.sync.update_files(filepath)

    @property
    def files(self) -> Dict[str, str]:
        
        logfiles = {}
        if isinstance(self.aio, AsyncFilesystemLogger) and isinstance(self.sync, SyncFilesystemLogger):
            logfiles.update(self.aio.filepaths)
            logfiles.update(self.sync.filepaths)

        return logfiles




import platform
import os
from typing import List, Optional
from werkflow.modules.shell import Shell
from werkflow.modules.base import Module


class Python(Module):

    def __init__(self) -> None:
        super().__init__()

        self.version = platform.python_version()
        self._shell = Shell()

    async def set_python_version(
        self,
        version: str=None,
        skip_if_exists: bool=True,
        force_reinstall: bool=False
    ):
        if version is None:
            version = self.version

        pyenv_install_command = f'pyenv install {version}'

        if skip_if_exists:
            pyenv_install_command = f'pyenv install -s {version}'

        if force_reinstall:
            pyenv_install_command = f'pyenv install -f {version}'

        await self._shell.run(
            pyenv_install_command, 
            silent=True
        )

        await self._shell.run(
            f'pyenv global {version}', 
            silent=True
        )

        await self._shell.eval('pyenv init -')

    async def create_venv(
        self, 
        environment_name: str
    ) -> str:

        absolute_path = await self._shell.to_absolute_path(environment_name)
        await self._shell.run(
            f'python -m venv {absolute_path}',
            silent=True
        )

        return os.path.join(
            absolute_path,
            'bin',
            'activate'
        )

    async def install_requirements(
        self,
        requirements: List[str],
        virtual_environment: str=None
    ) -> None:
        python_requirements = ','.join(requirements)

        if virtual_environment:
            virtual_environment_path = await self._shell.to_absolute_path(
                virtual_environment
            )

            await self._shell.run(
                f'source {virtual_environment_path} && pip install {python_requirements}',
                as_shell=True,
                silent=True
            )

        else:  
            await self._shell.run(
                f'pip install {python_requirements}',
                silent=True
            )

    async def install_requirements_from_file(
        self,
        install_requirements: Optional[List[str]]=None,
        filepath: str='requirements.txt',
        virtual_environment: str=None
    ) -> None:
        
        if install_requirements and len(install_requirements) > 0:

            requirements = '\n'.join(install_requirements)

            await self._shell.pipe_to_file(
                filepath,
                requirements,
                silent=True
            )

        else:
            await self._shell.touch(
                filepath,
                silent=True
            )

        if virtual_environment:

            virtual_environment_path = await self._shell.to_absolute_path(virtual_environment)

            return await self._shell.run(
                f'source {virtual_environment_path} && pip install -r {filepath}',
                as_shell=True,
                silent=True
            )
        
        else:

            return await self._shell.run(
                f'pip install -r {filepath}',
                silent=True
            )

    async def activate_virtual_environment(
        self,
        virtual_envrionment_path: str
    ) -> str:
        python_venv_absolute_location = await self._shell.to_absolute_path(virtual_envrionment_path)
        python_venv_activate_location = os.path.join(
            python_venv_absolute_location,
            'bin',
            'activate'
        )

        await self._shell.source(
            filepath=python_venv_activate_location
        )

        return python_venv_activate_location
    
    async def close(self):
        await self._shell.close()

    def abort(self):
        self._shell.abort()
        
    


from werkflow.modules.system.architecture.types import ArchitectureType
from .base import OperatingSystem
from .types import OperatingSystemType

class Windows(OperatingSystem):

    def __init__(self) -> None:
        super().__init__()

        self.name = 'Windows'
        self.type = OperatingSystemType.WINDOWS

        self.supported_architectures = {
            ArchitectureType.x86_64: 'x86_64',
            ArchitectureType.AMD_64: 'AMD64'
        }



from typing import List


class InvalidPromptError(Exception):

    def __init__(
        self, 
        invalid_prompt_name: str, 
        existing_prompt_names: List[str]
    ) -> None:
        
        valid_prompt_names = '\n\t-'.join(existing_prompt_names)

        super().__init__(
            f'Interactive Prompt {invalid_prompt_name} not registered.\nRegistered prompts include:\n\t-{valid_prompt_names}'
        )


class StepTimeoutError(Exception):

    def __init__(
        self, 
        step_name: str, 
        workflow_name: str,
        timeout: float
    ) -> None:
        super().__init__(
            f'Step - {step_name} - for workflow - {workflow_name} - timed out after - {timeout} - seconds.'
        )


import psutil
from psutil._common import sdiskpart
from .data_amount import DataAmount


class Drive:

    def __init__(self, drive_stats: sdiskpart) -> None:
        self.path = drive_stats.mountpoint
        self.name = drive_stats.device
        self.type = drive_stats.fstype
        self.options = drive_stats.opts

        disk_usage = psutil.disk_usage(self.path)

        self.total = DataAmount(disk_usage.total)
        self.used = DataAmount(disk_usage.used)
        self.available = DataAmount(disk_usage.free)
        self.percentage_used = disk_usage.percent

    @property
    def mounted(self) -> bool:
        return self.path is not None

    def update(self):
        disk_usage = psutil.disk_usage(self.path)

        self.used = DataAmount(disk_usage.used)
        self.available = DataAmount(disk_usage.free)
        self.percentage_used = disk_usage.percent

    @property
    def current_percentage_used(self):
        self.update()
        return self.percentage_used
    
    @property
    def current_available(self):
        self.update()
        return self.available
    
    @property
    def current_used(self):
        self.update()
        return self.used
        


import functools
from typing import List, Callable, Dict, Any
from werkflow.hooks.types.base.registrar import registrar
from werkflow.hooks.types.base.hook_types import HookType
from werkflow.prompt.types.base.base_prompt import BasePrompt
from .validator import StepHookValidator

@registrar(HookType.STEP)
def step(
    *names, 
    prompts: List[BasePrompt]=[], 
    skip_on_fail: bool=False,
    condition: Callable[[
        Dict[str, Any]
    ], bool]=None
):

    StepHookValidator(
        names=names,
        prompts=prompts,
        skip_on_fail=skip_on_fail,
        condition=condition
    )

    def wrapper(func):

        @functools.wraps(func)
        def decorator(*args, **kwargs):
            func(*args, **kwargs)
        
        return decorator
    return wrapper


from typing import Dict, Union, Optional
from werkflow.modules.system import System
from werkflow.modules.system.operating_system import OperatingSystemType
from werkflow.modules.base import Module
from .types import ZSHShell


class Shell(Module):

    def __init__(self) -> None:
        super().__init__()
        system = System()

        self._shell_types = {
            OperatingSystemType.MAC_OS: ZSHShell
        }

        self._shell = self._shell_types.get(
            system.os.type,
            ZSHShell
        )()

    def path_is_file(
        self,
        path: str
    ):
        return self._shell.path_is_file(path)
    
    def path_is_directory(
        self,
        path: str
    ):
        return self._shell.path_is_directory(path)
    
    async def get_envar(
        self,
        envar_name: str
    ):
        return await self._shell.get_envar(envar_name)
    
    async def set_envar(
        self,
        envar_name: str,
        envar_value: str
    ):
        await self._shell.set_envar(
            envar_name,
            envar_value
        )

    async def check_if_is_file(
        self,
        path: str
    ):
        return await self._shell.check_if_is_file(path)
    
    async def check_if_is_directory(
        self,
        path: str
    ):
        return await self._shell.check_if_is_directory(path)

    async def get_current_directory(self):
        return await self._shell.get_current_directory()

    async def to_absolute_path(
        self,
        filepath: str
    ):
        return await self._shell.to_absolute_path(
            filepath
        )
    
    async def to_relative_path(
        self,
        base_path: str,
        absolute_path: str
    ):
        return await self._shell.to_relative_path(
            base_path,
            absolute_path
        )
    
    async def change_directory(
        self,
        path: str
    ):
        await self._shell.change_directory(path)

    async def assume_admin(
        self, 
        password: str,
        silent: bool=False
    ):
        self._shell.silent = silent
        return await self._shell.assume_admin(password=password)
    
    async def run(
        self, 
        command: str, 
        command_input: str=None, 
        path: Optional[str]=None,
        env: Dict[str, str]={},
        as_shell: bool=False,
        silent: bool=False
    ):
        self._shell.silent = silent
        return await self._shell.run(
            command,
            command_input=command_input,
            path=path,
            env=env,
            as_shell=as_shell
        )
    
    async def copy(
        self,
        source_directory: str,
        destination_directory: str,
        recursive: bool=False,
        silent: bool=False
    ):
        self._shell.silent = silent
        await self._shell.copy(
            source_directory,
            destination_directory,
            recursive=recursive
        )
    
    async def move(
        self,
        source_directory: str,
        destination_directory: str,
        silent: bool=False
    ):
        self._shell.silent = silent
        await self._shell.move(
            source_directory,
            destination_directory
        )
    
    async def eval(
        self,
        command: str,
        env: Dict[str, str]={},
        silent: bool=False
    ) -> None:
        self._shell.silent = silent
        await self._shell.eval(
            command,
            env=env
        )
    
    async def touch(
        self,
        filepath: str,
        overwrite: bool=False,
        silent: bool=False
    ):
        self._shell.silent = silent
        await self._shell.touch(
            filepath,
            overwrite=overwrite
        )
    
    async def append_to_envfile(
        self,
        variable_name: str,
        variable_value: str,
        silent: bool=False
    ):
        self._shell.silent = silent
        return await self._shell.append_to_envfile(
            variable_name,
            variable_value
        )
    
    async def check_path_exists(
        self,
        filepath: str,
        silent: bool=False
    ):
        self._shell.silent = silent
        return await self._shell.check_path_exists(filepath)
    
    async def create_file(
        self,
        filepath: str,
        overwrite: bool=False,
        silent: bool=False
    ):
        self._shell.silent = silent
        return await self._shell.create_file(
            filepath,
            overwrite=overwrite
        )
    
    async def open_or_create_file(
        self,
        filepath: str,
        overwrite: bool=False,
        silent: bool=False
    ):
        self._shell.silent = silent
        return await self._shell.open_or_create_file(
            filepath,
            overwrite=overwrite
        )
    
    async def open_existing_file(
        self, 
        filepath: str,
        silent: bool=False
    ):
        self._shell.silent = silent
        return await self._shell.open_existing_file(filepath)
    
    async def pipe_to_file(
        self,
        filepath: str,
        data: Union[str, bytes],
        overwrite: bool=False,
        silent: bool=False
    ):
        self._shell.silent = silent
        return await self._shell.pipe_to_file(
            filepath,
            data,
            overwrite=overwrite
        )
    
    async def read_file(
            self,
            filepath: str,
            silent: bool=False
    ) -> str:
        self._shell.silent = silent
        return await self._shell.read_file(
            filepath
        )
    
    async def read_file_as_bytes(
        self,
        filepath: str,
        silent: bool=False
    ) -> bytes:
        self._shell.silent = silent
        return await self._shell.read_file(
            filepath,
            is_bytes=True
        )
    
    async def create_directory(
        self,
        path: str
    ):
        return await self._shell.create_directory(path)
    
    async def source(
        self,
        filepath: str=None,
        silent: bool=False
    ):
        self._shell.silent = silent
        return await self._shell.source(
            filepath=filepath
        )
    
    async def rm(
        self,
        path: str,
        pattern: Optional[str]=None,
        recurse: bool=False,
        silent: bool=False
    ):
        self._shell.silent = silent
        return await self._shell.rm(
            path,
            pattern=pattern,
            recurse=recurse
        )
    
    async def close(self):
        await self._shell.close()

    def abort(self):
        self._shell.abort()



import inspect
from collections import OrderedDict
from typing import List, Type, Dict
from .workflow import Workflow


class WorkflowGroup:
    workflows: List[Type[Workflow]]=[]

    def __init__(self) -> None:
        self.group_workflows: Dict[str, Workflow] = OrderedDict()
        self.workflow_names = []

        for workflow in self.workflows:
            new_workflow = type(workflow())()
            workflow_name = new_workflow.__class__.__name__

            new_workflow.__class__.__base__()
            
            self.group_workflows[workflow_name] = new_workflow
            self.workflow_names.append(workflow_name)

    def add_workflow(self, workflow: Type[Workflow]):
        new_workflow = workflow()

        new_workflow: Workflow = type(
                new_workflow.__class__.__name__, 
                (workflow, ), 
                dict(new_workflow.__dict__)
        )()
        
        new_workflow.__class__.__base__()

        workflow_name = new_workflow.__class__.__name__

        self.group_workflows[workflow_name] = new_workflow
        self.workflow_names.append(workflow_name)


from pydantic import BaseModel, StrictStr, StrictBool
from werkflow.prompt.types.base.base_prompt import BasePrompt
from typing import Tuple, List, Optional, Callable, Dict, Any


class StepHookValidator(BaseModel):
    names: Tuple[StrictStr, ...]
    prompts: List[BasePrompt]=[]
    skip_on_fail: StrictBool
    condition: Optional[
        Callable[
            [Dict[str, Any]],
            bool
        ]
    ]=None

    class Config:
        arbitrary_types_allowed=True




import uuid
import inspect
from typing import (
    Callable, 
    Any, 
    Tuple, 
    List,
    Dict
)
from werkflow.prompt.types.base.base_prompt import BasePrompt
from .hook_types import HookType


class BaseHook:

    def __init__(
        self,
        name: str,
        shortname: str,
        call: Callable[..., Any],
        *names: Tuple[str, ...],
        hook_type: HookType=None,
        prompts: List[BasePrompt]=[],
        skip_on_fail: bool=False,
        condition: Callable[
            [Dict[str, Any]],
            bool
        ]=None
    ) -> None:
        self.id = str(uuid.uuid4())
        self.name = name
        self.shortname = shortname
        self._call = call
        self.names: List[str] = list(names)
        self.args = inspect.signature(call)
        self.params = self.args.parameters
        self.type = hook_type
        self.skip_on_fail = skip_on_fail
        self.namespace: str = None
        self.workflow: str = None
        self.prompts = prompts
        self.condition = condition

    async def call(self, **kwargs):

        hook_args = {
            name: value for name, value in kwargs.items() if name in self.params
        }
        
        if self.condition and self.condition(kwargs):
            result = await self._call(**hook_args)

        elif self.condition is None:
            result = await self._call(**hook_args)

        else:
            result = {}

        if isinstance(result, dict):
            return {
                **kwargs,
                **result
            }
        
        return {
            **kwargs,
            self.shortname: result
        }




class MultipleWorkflowsViolationError(Exception):

    def __init__(
        self, 
        workflows_found: int,
        workflow_filepath: str
    ) -> None:
        super().__init__(
            f'Found {workflows_found} workflows in file or named workflow {workflow_filepath}.\nOnly one Workflow is allowed per file.'
        )


from pydantic import (
    BaseModel,
    validator
)
from werkflow.prompt.types.base.prompt_type import PromptType
from typing import Type


class ConfirmationPromptValidator(BaseModel):
    prompt_data_type: Type[str]
    prompt_type: PromptType

    class Config:
        arbitrary_types_allowed=True

    @validator('prompt_type')
    def validate_type(cls, val):
        assert val == PromptType.CONFIRMATION

        return val




import click
import os
from typing import List, Union, Optional, Callable, Any
from importlib.metadata import version
from art import text2art
from werkflow.logging import WerkflowLogger


class CLI(click.MultiCommand):

    command_files = {
        'setup': 'setup.py',
        'workflow': 'workflow.py'
    }
    logger = WerkflowLogger()

    def __init__(
        self, 
        name: Optional[str] = None, 
        invoke_without_command: bool = False, 
        no_args_is_help: Optional[bool] = None, 
        subcommand_metavar: Optional[str] = None, 
        chain: bool = False, 
        result_callback: Optional[Callable[..., Any]] = None, 
        **attrs: Any
    ) -> None:
        super().__init__(
            name, 
            invoke_without_command, 
            no_args_is_help, 
            subcommand_metavar, 
            chain, 
            result_callback, 
            **attrs
        )

        self.logger.initialize()

        header_text = text2art('werkflow', font='rozzo').strip('\n')
        werkflow_version = version('werkflow')

        self.logger.console.sync.info(f'\n{header_text} v{werkflow_version}\n\n')


    def list_commands(self, ctx: click.Context) -> List[str]:
        rv = []
        for filename in self.command_files.values():
            rv.append(filename[:-3])
        rv.sort()
        return rv

    def get_command(self, ctx: click.Context, name: str) -> Union[click.Command, None]:

        ns = {}
        
        command_file = os.path.join(
            os.path.dirname(__file__),
            self.command_files.get(name)
        )

        with open(command_file) as f:
            code = compile(f.read(), command_file, 'exec')
            eval(code, ns, ns)
        
        return ns.get(name)


from werkflow.modules.base import Module


class MissingModuleError(Exception):

    def __init__(
        self,
        module: Module
    ): 
        module_name = module.__class__.__name__
        dependencies = module.dependencies

        dependency_names = '\n\t-'.join(dependencies)

        super().__init__(
            f'Module {module_name} is missing or its dependencies:\n\t-{dependency_names}\nhave not been installed.'
        )


import click


@click.group(help="Commands to setup global developer environment or a project.")
def setup():
    pass


@setup.command(help='Install custom dependencies or environment')
@click.argument('name')
@click.option(
    '--type',
    default='python',
    help='Resource type to use for installation.'
)
@click.option(
    '--env-version',
    default='latest',
    help='Version of language/resource to use.'
)
@click.option(
    '--using',
    default='',
    help='Comma-delimited list of dependencies/requirements.'
)
def install(
    name: str,
    type: str,
    env_version: str,
    using: str
):
    dependencies = using.split(',')


import sys
import logging
from typing import Mapping, Any, List
from aiologger.levels import LogLevel
from .logger_types import LoggerTypes


class SyncLogger(logging.Logger):

    def __init__(
        self, 
        logger_name: str=None, 
        logger_type: LoggerTypes=LoggerTypes.CONSOLE,
        log_level: LogLevel = LogLevel.INFO, 
        logger_enabled: bool=True
    ) -> None:
        super().__init__(logger_name, level=log_level)
        self.logger_name = logger_name
        self.logger_type = logger_type
        self.log_level = log_level
        self.logger_enabled = logger_enabled
        self.handler = None

    def initialize(self, pattern: str, datefmt_pattern: str=None):
        stream_handler = logging.StreamHandler(stream=sys.stdout)
        self.setLevel(self.level)
        stream_handler.setFormatter(
            logging.Formatter(pattern, datefmt=datefmt_pattern)
        )

        self.addHandler(stream_handler)


    def _log(self, level: int, msg: object, *args: List[Any], **kwargs: Mapping[str, Any]) -> None:
        if self.logger_enabled:
            return super()._log(level, msg, *args, **kwargs)


import asyncio
import networkx
import inspect
import functools
import click
import os
from typing import Dict, List, Any
from dotenv import dotenv_values
from werkflow.hooks.types.base.base_hook import BaseHook
from werkflow.hooks.types.base.registrar import registrar
from werkflow.logging import WerkflowLogger
from .workflow_group import WorkflowGroup
from .workflow import Workflow


class Graph:

    def __init__(
        self,
        workflows: WorkflowGroup,
        no_prompt: bool=False,
        werkflow_config: Dict[str, Any]={}
    ) -> None:

        self.logger = WerkflowLogger()
        self.logger.initialize()

        self._no_prompt = no_prompt
        self._werkflow_config = werkflow_config
        self._project_options: Dict[str, Any] = self._werkflow_config.get(
            'project_options', {}
        )

        self._env = {}
        
        dot_env_path = f'{os.getcwd()}/.env'
        if os.path.exists(dot_env_path):
            self._env = dotenv_values(dot_env_path)

        self._project_options.update({
            key.lower(): value for key, value in self._env.items()
        })

        self._project_options['command_directory'] = os.getcwd()

        self._graphs: Dict[str, networkx.DiGraph] = {}
        self._execution_orders: Dict[str, List[List[str]]] = {}
        self._workflows: Dict[str, Workflow] = workflows.group_workflows
        self._workflow_hooks: Dict[str, Dict[str, BaseHook]] = {}

    def setup(self):

        for workflow in self._workflows.values():

            workflow.werkflow_config.update(self._werkflow_config)

            workflow_graph = networkx.DiGraph()
            workflow_hooks = {}
            workflow_name = workflow.__class__.__name__

            workflow_methods = inspect.getmembers(
                workflow, 
                predicate=inspect.ismethod
            )

            for _, method in workflow_methods:
                hook = registrar.all.get(method.__qualname__)

                if hook:
                    hook.workflow = workflow_name

                    hook._call = hook._call.__get__(workflow, workflow.__class__)
                    setattr(workflow, hook.shortname, hook._call)

                    workflow_hooks[hook.shortname] = hook

            workflow_graph.add_nodes_from([
                (
                    hook_name, 
                    {"hook": hook}
                ) for hook_name, hook in workflow_hooks.items()
            ])
            
            for hook in workflow_hooks.values():
                for dependency in hook.names:
                    if workflow_graph.nodes.get(dependency):
                        workflow_graph.add_edge(dependency, hook.shortname)

            execution_order: List[List[str]] = [
                generation for generation in networkx.topological_generations(workflow_graph)
            ]

            self._execution_orders[workflow_name] = execution_order
            self._workflow_hooks[workflow_name] = workflow_hooks
            self._graphs[workflow_name] = workflow_graph

    async def run(self):

        next_args: Dict[str, Any] = dict(self._project_options)

        loop = asyncio.get_running_loop()

        for workflow_name, execution_order in self._execution_orders.items():
            
            workflow = self._workflows.get(workflow_name)
            workflow_hooks = self._workflow_hooks.get(workflow_name)


            for generation in execution_order:
                generation_hooks: List[BaseHook] = [
                    workflow_hooks.get(hook_name) for hook_name in generation
                ]

                current_steps = ', '.join(
                    list(set([step.shortname for step in generation_hooks]))
                )

                if self.logger.spinner.logger_enabled:
                    await self.logger.spinner.append_message(f"Executing steps - {current_steps}")

                generation_prompts = []
                hooks_with_prompts = []
                for hook in generation_hooks:
                    hook_prompts = [hook for hook in generation_hooks if len(hook.prompts) > 0]
                    generation_prompts.extend(hook_prompts)

                    if len(hook_prompts) > 0:
                        hooks_with_prompts.append(hook.name)

                for hook in generation_hooks:
                    for prompt in hook.prompts:

                        result_key = prompt.result_key
                        if result_key is None:
                            result_key = hook.shortname

                        result_env_key = result_key.upper()
                        if self._no_prompt:
                            result = self._project_options.get(prompt.result_key)

                            if result is None:
                                result = await loop.run_in_executor(
                                    None,
                                    functools.partial(
                                        os.getenv,
                                        result_env_key
                                    )
                                )

                            if result is not None:
                                next_args[result_key] = result
                            
                        skipped = prompt.skipped
                        if prompt.condtition:
                            skipped = prompt.condtition(next_args) is False

                        if skipped is False and self._no_prompt is False:

                            await loop.run_in_executor(
                                None,
                                functools.partial(
                                    click.echo,
                                    ''
                                )
                            )

                            result = await prompt.ask()
                            await prompt.confirm(result)
                            next_args[result_key] = result

                            prompt.close()

                if self.logger.spinner.logger_enabled:
                    async with self.logger.spinner as status_spinner:
                        results: List[Dict[str, Any]] = await asyncio.gather(*[
                            asyncio.create_task(hook.call(
                                **next_args
                            )) for hook in generation_hooks
                        ])

                        status_spinner.group_finalize()
                        await status_spinner.ok('✔')

                else:
                    results: List[Dict[str, Any]] = await asyncio.gather(*[
                        asyncio.create_task(hook.call(
                            **next_args
                        )) for hook in generation_hooks
                    ])

                for result in results:
                    next_args.update({
                        result_key: result_value for result_key, result_value in result.items() if result_value is not None
                    })

            await workflow.close()

       



from .types.confirmation.confirmation_prompt import ConfirmationPrompt
from .types.input.input_prompt import InputPrompt
from .types.key_value.key_value_prompt import KeyValuePrompt
from .types.option.option_prompt import OptionPrompt
from .types.secure.secure_prompt import SecurePrompt
from .types.repeat.repeat_prompt import RepeatPrompt


from pydantic import (
    BaseModel,
    StrictBool,
    StrictStr
)
from typing import (
    Union, 
    Optional,
    Callable,
)


class BasePromptValidator(BaseModel):
    prompt_message: Optional[StrictStr]
    prompt_skipped: StrictBool
    prompt_result_key: Optional[str]
    prompt_default: Optional[Union[int, float, str, bool]]
    prompt_condition: Optional[Callable[..., bool]]
    prompt_confirmation_message: Optional[Union[Callable[..., StrictStr], StrictStr]]




import asyncio
import functools
import click
from concurrent.futures import ThreadPoolExecutor
from werkflow.modules.system import System
from termcolor import colored
from typing import Optional, Any, Callable, Union, Dict
from .base_prompt_validator import BasePromptValidator


class BasePrompt:

    def __init__(
        self,
        message: str,
        skip: bool=False,
        result_key: str=None,
        default: Optional[Any]=None,
        condition: Optional[Callable[..., bool]] = None, 
        confirmation_message: Optional[Union[str, Callable[..., str]]] = None, 
    ) -> None:

        system = System()

        self._loop: asyncio.AbstractEventLoop = None
        self._executor = ThreadPoolExecutor(
            max_workers=system.configuration.cores.physical
        )

        self.prompt_color = 'cyan'
        self.prompt_frame = colored('●', color=self.prompt_color)
        self.confirmation_frame = colored('✔', color=self.prompt_color)

        validated_prompt = BasePromptValidator(
            prompt_message=message,
            prompt_skipped=skip,
            prompt_result_key=result_key,
            prompt_default=default,
            prompt_condition=condition,
            prompt_confirmation_message=confirmation_message
        )

        self.message = validated_prompt.prompt_message
        self.skipped = validated_prompt.prompt_skipped
        self.result_key = validated_prompt.prompt_result_key
        self.default = validated_prompt.prompt_default
        self.condtition = validated_prompt.prompt_condition
        self.confirmation_message = validated_prompt.prompt_confirmation_message

    async def ask(self):
        raise NotImplementedError('Ask method not implemented on base Prompt class.')

    async def confirm(self, response: Any):
        if self.confirmation_message:
            
            if isinstance(self.confirmation_message, str):
                confirmation_text = self.confirmation_message

            else:
                confirmation_text = self.confirmation_message(response)

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    click.echo,
                    f'{self.confirmation_frame} {confirmation_text}',
                )
            )

    def close(self):
        self._executor.shutdown()


from werkflow.prompt.types.base.base_prompt import BasePrompt
from werkflow.prompt.types.base.prompt_type import PromptType
from pydantic import BaseModel, StrictStr, StrictBool, validator
from typing import Callable, Optional, Any


class KeyValuePromptValidator(BaseModel):
    key_prompt: Any
    value_prompt: Any
    prompt_result_key: Optional[StrictStr]
    prompt_type: PromptType
    prompt_condition: Optional[Callable[..., StrictBool]]

    class Config:
        arbitrary_types_allowed=True

    @validator('key_prompt')
    def validate_key_prompt(cls, val):
        assert issubclass(type(val), BasePrompt)
        
        return val

    @validator('value_prompt')
    def validate_value_prompt(cls, val):
        assert issubclass(type(val), BasePrompt)

        return val

    @validator('prompt_type')
    def validate_type(cls, val):
        assert val == PromptType.KEY_VALUE

        return val



import asyncio
import click
import os
import json
from werkflow.cli.signals import add_abort_handler
from werkflow.cli.import_tools.workflow import import_workflow
from werkflow.graph import (
    Graph,
    WorkflowGroup
)
from werkflow.logging import (
    WerkflowLogger,
    LoggerTypes,
    logging_manager
)
from werkflow.modules.base import Module
from typing import List


@click.group(help='Commands to run and manage workflows.')
def workflow():
    pass

@workflow.command(
    help='Run the workflow at the given path or by the given name.'
)
@click.option(
    '--ci',
    is_flag=True,
    help='Mute CLI graphics for CI compatability.'
)
@click.option(
    '--no-prompt',
    is_flag=True,
    help='Mute all interactive prompts and default to values in .werkflow.json.'
)
@click.option(
    '--config-path',
    show_default=True,
    default=f'{os.getcwd()}/.werkflow.json',
    help='Path to existing .werkflow.json.'
)
@click.option(
    '--log-level',
    default='info',
    help='Set log level.'
)
@click.option(
    '--logfiles-directory',
    show_default=True,
    default=f'{os.getcwd()}/logs',
    help='Output directory for logfiles. If the directory does not exist it will be created.'
)
@click.argument('path')
def run(
    ci: bool,
    no_prompt: bool,
    config_path: str,
    path: str,
    log_level: str,
    logfiles_directory: str
):
    werkflow_config = {}
    if os.path.exists(config_path):
        with open(config_path) as werkflow_config_file:
            werkflow_config = json.load(werkflow_config_file)

    werkflow_config['config_path'] = config_path

    disabled_loggers = [
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.DISTRIBUTED_FILESYSTEM
    ]

    if ci:
        disabled_loggers.extend([
            LoggerTypes.SPINNER,
            LoggerTypes.WERKFLOW
        ])

    logging_manager.disable(*disabled_loggers)

    logging_manager.update_log_level(log_level)
    logging_manager.logfiles_directory = logfiles_directory

    if os.path.exists(logfiles_directory) is False:
        os.mkdir(logfiles_directory)

    elif os.path.isdir(logfiles_directory) is False:
        os.remove(logfiles_directory)
        os.mkdir(logfiles_directory)

    logger = WerkflowLogger()
    logger.initialize()

    discovered = import_workflow(path)
    workflow: WorkflowGroup = discovered.get('workflow')
    modules: List[Module] = discovered.get('modules')

    module_names = [module.__class__.__name__ for module in modules]
    
    logger['console'].sync.info(f'Loading graph - {workflow.__class__.__name__}')
    
    if len(modules) > 0:
        modules_loaded = ', '.join(module_names)
        logger['console'].sync.info(f'Using modules - {modules_loaded}')

    try:
        loop = asyncio.get_event_loop()

    except Exception:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    
    logger.filesystem.sync.create_logfile('werkflow.core.log')
    logger.filesystem.create_filelogger('werkflow.core.log')

    graph = Graph(
        workflow,
        no_prompt=no_prompt,
        werkflow_config=werkflow_config
    )

    graph.setup()
    
    add_abort_handler(
        loop,
        logger,
        graph
    )

    loop.run_until_complete(graph.run())    

    if logger.spinner.logger_enabled:
        logger.console.sync.info(f'\nGraph - {workflow.__class__.__name__} - completed! {graph.logger.spinner.display.total_timer.elapsed_message}\n')

    else:
        logger.console.sync.info(f'\nGraph - {workflow.__class__.__name__} - completed!\n')


import psutil
from .data_amount import DataAmount

class Memory:

    def __init__(self) -> None:
        memory_stats = psutil.virtual_memory()
    
        self.total = DataAmount(memory_stats.total)
        self.used = DataAmount(memory_stats.used)
        self.available = DataAmount(memory_stats.available)
        self.percentage_used = memory_stats.percent

    def update(self):
        memory_stats = psutil.virtual_memory()
        self.used = DataAmount(memory_stats.used)
        self.available = DataAmount(memory_stats.available)
        self.percentage_used = memory_stats.percent

    @property
    def current_percentage_used(self):
        self.update()
        return self.percentage_used
    
    @property
    def current_available(self):
        self.update()
        return self.available
    
    @property
    def current_used(self):
        self.update()
        return self.used
        

    



from typing import Dict, Type
from werkflow.hooks.types.step.hook import StepHook
from .base_hook import BaseHook
from .hook_types import HookType


class Registrar:
    all: Dict[str, BaseHook]={}
    module_paths: Dict[str, str] = {}

    def __init__(self, hook_type: HookType) -> None:
        self._hook_type = hook_type
        self._hook_types_map = {
            HookType.STEP: lambda *args, **kwargs: StepHook(
                *args,
                **kwargs
            )
        }

    def __call__(self, hook):
        hook = self._hook_types_map.get(self._hook_type)
        self.module_paths[hook.__name__] = hook.__module__

        def wrap_hook(*args, **kwargs):

            def wrapped_method(func):

                hook_name = func.__qualname__
                hook_shortname = func.__name__

                hook = self._hook_types_map[self._hook_type]

                hook_args = args
                args_count = len(args)
                
                if args_count < 1:
                    hook_args = []

                self.all[hook_name] = hook(
                    hook_name,
                    hook_shortname,
                    func,
                    *hook_args,
                    **kwargs
                )
                
                return func
            
            return wrapped_method

        return wrap_hook



def makeRegistrar() -> Type[Registrar]:
    return Registrar

registrar = makeRegistrar()


import re
from datetime import timedelta

class TimeParser:

    def __init__(self, time_amount: str) -> None:
        self.UNITS = {
            's':'seconds', 
            'm':'minutes', 
            'h':'hours', 
            'd':'days', 
            'w':'weeks'
        }
        self.time = float(
            timedelta(
                **{
                    self.UNITS.get(
                        m.group(
                            'unit'
                        ).lower(), 
                        'seconds'
                    ): float(m.group('val')
                )
                    for m in re.finditer(
                        r'(?P<val>\d+(\.\d+)?)(?P<unit>[smhdw]?)', 
                        time_amount, 
                        flags=re.I
                    )
                }
            ).total_seconds()
        )


import click
from pydantic import (
    BaseModel,
    StrictBool,
    StrictStr,
    StrictInt,
    validator
)
from typing import Union, Type, TypeVar
from werkflow.prompt.types.base.prompt_type import PromptType

T = TypeVar('T')


class SecurePromptValidator(BaseModel):
    prompt_data_type: Type[T]
    prompt_type: PromptType

    class Config:
        arbitrary_types_allowed=True

    @validator('prompt_type')
    def validate_type(cls, val):
        assert val == PromptType.SECURE

        return val




from werkflow.prompt.types.base.base_prompt import BasePrompt
from werkflow.prompt.types.base.prompt_type import PromptType
from pydantic import BaseModel, StrictBool, validator
from typing import Callable, Any, Union


class RepeatPromptValidator(BaseModel):
    prompt_repeated: object
    prompt_break_condition: Callable[..., StrictBool]
    prompt_type: PromptType

    class Config:
        arbitrary_types_allowed=True

    @validator('prompt_repeated')
    def validate_prompt(cls, val):
        assert issubclass(type(val), BasePrompt)

        return val

    @validator('prompt_type')
    def validate_type(cls, val):
        assert val == PromptType.REPEAT

        return val



import asyncio
import orjson
from collections import deque
from typing import (
    Callable, 
    Coroutine, 
    Any, 
    Tuple, 
    List, 
    Deque,
    Union,
    Dict,
    Optional
)
from werkflow.modules import (
    File,
    Shell,
    System
)
from werkflow.logging import WerkflowLogger
from .exceptions import StepTimeoutError

class Workflow:
    priority=1

    def __init__(self) -> None: 
        self.file = File()
        self.system = System()
        self.shell = Shell()

        self.pending: Deque[asyncio.Task] = deque()
        self.logger = WerkflowLogger()
        self.logger.initialize()
        self.werkflow_config: Dict[str, Any] = {}

    def get_project_option(
        self,
        option_name: str
    ): 
        options: Dict[str, Any] = self.werkflow_config.get('project_options')
        return options.get(option_name)

    async def update_project_options(
        self,
        option_name: str,
        data: Any
    ):
        options: Dict[str, Any] = dict(self.werkflow_config.get('project_options'))
        options[option_name] = data

        self.werkflow_config['project_options'] = options

        path: str = self.werkflow_config.get('config_path')
        
        current_config: str = await self.shell.read_file(
            path,
            silent=True
        )

        current_config: Dict[str, Any] = orjson.loads(current_config)
        current_config.update(self.werkflow_config)

        await self.shell.pipe_to_file(
            path,
            orjson.dumps(
                current_config,
                option=orjson.OPT_INDENT_2
            ).decode(),
            overwrite=True,
            silent=True
        )

    async def as_async(
        self,
        call: Callable[..., Any],
        *args: Tuple[Any, ...],
        **kwargs: Dict[str, Any]
    ):
        return await asyncio.to_thread(
            call,
            *args,
            **kwargs
        )

    async def sequence(
        self,
        *jobs: Tuple[
            Coroutine[
                None,
                None,
                Any
            ]
        ]
    ):
        results: List[Any] = []

        for job in jobs:
            results.append(
                await job
            )

        return results

    async def batch(
        self,
        *jobs: Tuple[
            Coroutine[
                None,
                None,
                Any
            ]
        ],
        timeout: Optional[
            Union[int, float]
        ]=None
    ) -> List[Any]:

        completed_jobs: List[Any] = await asyncio.wait_for(
            asyncio.gather(*[
                job for job in jobs
            ]),
            timeout=timeout
        )

        results = []

        for completed_job in completed_jobs:

            if isinstance(completed_job, list):
                results.extend(completed_job)

            else:
                results.append(completed_job)
        
        return results

    async def finalize(
        self,
        timeout: float=None
    ) -> List[Any]:
        results = []
        for result in asyncio.as_completed(self.pending, timeout=timeout):
            results.append(await result)

        return results

    def defer(
        self,
        call: Callable[..., Coroutine[None, None, Any]],
        *args: Tuple[Any, ...],
        timeout: float=60,
        fail_on_timeout: bool = True,
        **kwargs: Dict[str, Any],

    ) -> None:
        self.pending.append(
            asyncio.create_task(
                self.wait(
                    call(*args, **kwargs),
                    timeout=timeout,
                    fail_on_timeout=fail_on_timeout
                )
            )
        )
    
    async def wait(
        self,
        call: Callable[..., Coroutine[None, None, Any]],
        timeout: float=60,
        fail_on_timeout: bool = True
    ) -> Union[Any, StepTimeoutError]:
        try:
            return await asyncio.wait_for(call, timeout=timeout)

        except (asyncio.TimeoutError, asyncio.CancelledError,):
            error = StepTimeoutError(
                call.__name__,
                self.__class__.__name__,
                timeout
            )

            if fail_on_timeout:
                raise error
            
            return error

    async def close(self):
        await self.system.close()
        await self.shell.close()

    def abort(self):
        self.system.abort()
        self.shell.abort()


import click
from pydantic import (
    BaseModel,
    validator
)
from werkflow.prompt.types.base.prompt_type import PromptType


class OptionPromptValidator(BaseModel):
    prompt_data_type: click.Choice
    prompt_type: PromptType

    class Config:
        arbitrary_types_allowed=True

    @validator('prompt_type')
    def validate_type(cls, val):
        assert val == PromptType.OPTION

        return val




import ntpath
import importlib
import sys
from typing import Dict, Union
from werkflow.cli.exceptions import (
    MultipleWorkflowsViolationError,
    EmptyWorkflowFileError
)
from werkflow.graph import Workflow, WorkflowGroup
from werkflow.modules.base import Module
from pathlib import Path

def import_workflow(path: str) -> Dict[str, Union[Workflow, Module]]:
    package_dir = Path(path).resolve().parent
    package_dir_path = str(package_dir)
    package_dir_module = package_dir_path.split('/')[-1]
    
    package = ntpath.basename(path)
    package_slug = package.split('.')[0]
    spec = importlib.util.spec_from_file_location(f'{package_dir_module}.{package_slug}', path)
    
    if path not in sys.path:
        sys.path.append(str(package_dir.parent))

    module = importlib.util.module_from_spec(spec)

    sys.modules[module.__name__] = module

    spec.loader.exec_module(module)

    workflows = list({cls.__name__: cls for cls in Workflow.__subclasses__()}.values())
    workflow_groups = list({cls.__name__: cls() for cls in WorkflowGroup.__subclasses__()}.values())
    modules = list({cls.__name__: cls for cls in Module.__subclasses__()}.values())

    prioritized_workflows = list(sorted(
        workflows,
        key=lambda workflow: workflow.priority
    ))

    default_workflow_group = WorkflowGroup()

    if len(workflow_groups) > 0:
        for group in workflow_groups:

            for workflow in prioritized_workflows:
                workflow_name = workflow.__name__

                if workflow_name not in group.workflow_names:
                    default_workflow_group.add_workflow(workflow)

        if len(default_workflow_group.group_workflows) > 0 :
            workflow_groups.append(default_workflow_group)

    else:
        for workflow in prioritized_workflows:
            workflow_name = workflow.__class__.__name__
            default_workflow_group.add_workflow(workflow)

        if len(default_workflow_group.group_workflows) > 0 :
            workflow_groups.append(default_workflow_group)

    groups_discovered = len(workflow_groups)

    if groups_discovered < 1:
        raise EmptyWorkflowFileError(path)
    
    elif groups_discovered > 1:
        raise MultipleWorkflowsViolationError(
            groups_discovered,
            path
        )
    
    else:
        workflow_group: WorkflowGroup = workflow_groups.pop()
        return {
            'workflow': workflow_group,
            'modules': [
                module() for module in modules
            ]
        }



class EmptyWorkflowFileError(Exception):

    def __init__(self, path: str) -> None:
        super().__init__(
            f'Specified file or named workflow at {path} is empty.\n Please add a workflow.'
        )


import asyncio
import functools
import pathlib
from concurrent.futures import ThreadPoolExecutor
from werkflow.modules.base import Module
from werkflow.modules.shell import Shell
from werkflow.modules.system import System
from typing import List


class File(Module):

    def __init__(self) -> None:
        super().__init__()

        self._system = System()
        self._shell = Shell()

        self._loop: asyncio.AbstractEventLoop = None
        self._executor = ThreadPoolExecutor(
            max_workers=self._system.configuration.cores.physical
        )

    async def list_matching_files(
        self,
        path: str,
        pattern: str
    ) -> List[str]:
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        absolute_path = await self._shell.to_absolute_path(path)

        search_path = pathlib.Path(absolute_path)
        
        results = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                search_path.rglob,
                pattern
            )
        )

        return [
            str(result) for result in results
        ]
        

    async def close(self):
        await self._shell.close()
        await self._system.close()
        self._executor.shutdown()

    def abort(self):
        self._shell.abort()
        self._system.abort()
        self._executor.shutdown()


from typing import (
    Callable,
    Tuple,
    Any,
    List,
    Dict
)
from werkflow.hooks.types.base.base_hook import BaseHook
from werkflow.hooks.types.base.hook_types import HookType
from werkflow.prompt.types.base.base_prompt import BasePrompt


class StepHook(BaseHook):

    def __init__(
            self, name: str, 
            shortname: str, 
            call: Callable[..., Any], 
            *names: Tuple[str, ...],
            prompts: List[BasePrompt]=[],
            skip_on_fail: bool=False,
            condition: Callable[
                [Dict[str, Any]],
                bool
            ]=None
        ) -> None:

        super().__init__(
            name, 
            shortname,
            call, 
            *names,
            hook_type=HookType.STEP,
            prompts=prompts,
            skip_on_fail=skip_on_fail,
            condition=condition
        )



import psutil
from typing import Iterator, Union, Dict
from werkflow.modules.system.exceptions import MissingVolumeException
from .drive import Drive


class DriveGroup:

    def __init__(self) -> None:
        drives = psutil.disk_partitions()

        self._drives: Dict[str, Drive] = {
            drive.device: Drive(drive) for drive in drives
        }

    def __iter__(self) -> Iterator[Drive]:
        for drive in self._drives.values():
            yield drive

    def list(self) -> Iterator[str]:
        for drive_name in self._drives.keys():
            yield drive_name

    def __getitem__(self, drive_name) -> Drive:
        return self.get(drive_name)

    def get(self, drive_name: str) -> Drive:
        drive = self._drives.get(drive_name)
        if drive is None:
            raise MissingVolumeException(
                list(self._drives.keys())
            )
        
        return drive
    
    def exists(self, drive_name: str) -> bool:
        return self._drives.get(drive_name) is not None
    
    def all_like(self, search_drive_name: str) -> Iterator[Drive]:
        for drive_name, drive in self._drives.items():
            if search_drive_name in drive_name:
                yield drive
    
    def like(self, search_drive_name: str) -> Union[Drive, None]:

        for drive_name, drive in self._drives.items():
            if search_drive_name in drive_name:
                return drive
            
        return None
    
    def update(self) -> None:
        drives = psutil.disk_partitions()

        for drive in drives:
            if self._drives.get(drive.device) is None:
                self._drives[drive.device] = Drive(drive)

        for drive in self._drives.values():
            drive.update()

    def at_path(self, path: str) -> Drive:
        for drive in self._drives.values():
            if path == drive.path:
                return drive
            
        raise MissingVolumeException(
            list(self._drives.keys())
        )
    
    def at_path_like(self, path: str) -> Iterator[Drive]:
        for drive in self._drives.values():
            if path in drive.path:
                yield drive

    





import os

from setuptools import find_packages, setup

current_directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(current_directory, 'README.md'), "r") as readme:
    package_description = readme.read()

version_string = "0.1.4"
version_path = os.path.join(current_directory, ".version")
if os.path.exists(version_path):
    with open (os.path.join(current_directory, ".version"), 'r') as version_file:
        version_string = version_file.read()

setup(
    name="werkflow",
    version=version_string,
    description="Composable workflows for all.",
    long_description=package_description,
    long_description_content_type="text/markdown",
    author="Ada Lundhe",
    author_email="corpheus91@gmail.com",
    url="https://github.com/scorbettUM/werkflow",
    packages=find_packages(),
    keywords=[
        'pypi', 
        'cicd', 
        'python',
        'setup',
        'repo',
        'project',
        'migrate',
        'monorepo'
    ],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    install_requires=[
        'click',
        'psutil',
        'distro',
        'pydantic',
        'networkx',
        'aiologger',
        'yaspin',
        'art',
        'python-dotenv',
        'orjson'
    ],
    entry_points = {
        'console_scripts': [
            'werkflow=werkflow.cli:run',
        ],
    },
    extras_require = {
        'all': [
            'werkflow-aws',
            'werkflow-docker',
            'werkflow-encryption',
            'werkflow-git',
            'werkflow-github',
            'werkflow-http',
            'werkflow-metrik',
            'werkflow-secrets'
        ],
        'aws': [
            'werkflow-aws'
        ],
        'docker': [
            'werkflow-docker'
        ],
        'encryption': [
            'werkflow-encryption',
        ],
        'git': [
            'werkflow-git'
        ],
        'github': [
            'werkflow-github'
        ],
        'http': [
            'werkflow-http'
        ],
        'metrik': [
            'werkflow-metrik'
        ],
        'secrets': [
            'werkflow-secrets'
        ]
    },
    python_requires='>=3.8'
)



from werkflow.modules.base import Module
from pydantic import BaseModel, validator
from typing import Tuple, Type


class RequiresValidator(BaseModel):
    modules: Tuple[Type[Module], ...]

    class Config:
        allow_arbitrary_types=True

    @validator('modules')
    def validate(cls, val: Tuple[Type[Module], ...]):
        assert len(val) > 0
        
        for module_type in val:
            assert issubclass(module_type, Module), "Only Modules can be provided to the @requires hook."

            module_dependencies = f'\n\t-'.join(module_type.dependencies)

            assert module_type.module_enabled is True, f"The {module_type.__name__} Module is missing or its dependencies:\n\n\t-{module_dependencies}\n\n  have not been installed."

        return val


from typing import List


class UnsupportedArchitectureException(Exception):

    def __init__(
        self, 
        operating_system: str,
        provided_arch_option: str,
        supported_arch_options: List[str],
    ) -> None:

        supported_options = ', '.join(supported_arch_options)

        super().__init__(
            f'Unsupported architecture for {operating_system}.\n -Found: {provided_arch_option}\n- Supports: {supported_options}'
        )


from .missing_volume_exception import MissingVolumeException
from .unsupported_architecture_exception import UnsupportedArchitectureException
from .unsupported_operating_system_exception import UnsupportedOperatingSystemException



import asyncio
from typing import List
from aiologger.levels import LogLevel
from .spinner import ProgressText
from .logger_types import (
    LoggerTypes,
    LoggerTypesMap
)


class LoggingManager:

    def __init__(self) -> None:

        self.logger_types = LoggerTypesMap()
        self.enabled_loggers = {}

        for logger_type in self.logger_types.async_loggers.keys():
            self.enabled_loggers[logger_type] = True

        self.log_levels = {
            'info': LogLevel.INFO,
            'debug': LogLevel.DEBUG,
            'error': LogLevel.ERROR,
            'warning': LogLevel.WARNING,
            'warn': LogLevel.WARN,
            'critical': LogLevel.CRITICAL,
            'fatal': LogLevel.FATAL
        }

        self.log_level = LogLevel.INFO
        self.log_level_name = 'info'
        self.logfiles_directory = None
        self.progress_display = ProgressText()

    def update_log_level(self, log_level_name: str):
        self.log_level = self.log_levels.get(log_level_name, LogLevel.INFO)
        self.log_level_name = log_level_name

    def get_logger_enabled_state(self, logger_type: LoggerTypes):
        return self.enabled_loggers.get(logger_type)

    def enable(self, *logger_types: List[LoggerTypes]):
        for logger_type in logger_types:
            self.enabled_loggers[logger_type] = True

    def disable(self, *logger_types: List[LoggerTypes]):
        for logger_type in logger_types:
            self.enabled_loggers[logger_type] = False

    def list_enabled(self):
        return [
            logger_type for logger_type, logger_enabled in self.enabled_loggers.items() if logger_enabled
        ]

    def list_disabled(self):
        return [
            logger_type for logger_type, logger_enabled in self.enabled_loggers.items() if not logger_enabled
        ]

    def list_enabled_type_names(self):
        enabled_loggers = self.list_enabled()
        return [
            self.logger_types.get_name(logger_type) for logger_type in enabled_loggers
        ]

    def list_disabled_type_names(self):
        disabled_loggers = self.list_disabled()
        return [
            self.logger_types.get_name(logger_type) for logger_type in disabled_loggers
        ]


logging_manager = LoggingManager()


import os
import datetime
import signal
from typing import Any, List, Dict, Coroutine
from aiologger.levels import LogLevel
from werkflow.logging.logger_types.handlers.async_file_handler import RolloverInterval
from yaspin.spinners import Spinners
from werkflow.logging.logger_types.logger_types import LoggerTypes
from werkflow.logging.spinner import ProgressText


class LoggingConfig:
    logger_name: str=None
    logger_type: LoggerTypes=LoggerTypes.CONSOLE
    logfiles_directory: str=f'{os.getcwd()}/logs'
    log_level: LogLevel=LogLevel.INFO
    logger_enabled: bool=None
    filesystem_rotation_interval_type: RolloverInterval=RolloverInterval.DAYS
    filesystem_rotation_interval: int=1
    filesystem_backup_count: int=1
    filesystem_rotation_time: datetime.time=None
    spinner_type: Spinners=Spinners.bouncingBar
    spinner_color: str='cyan'
    spinner_on_color: str=None
    spinner_attrs: List[str]=["bold"]
    spinner_reversal: bool=False
    spinner_side: str="left"
    spinner_sigmap: Dict[signal.Signals, Coroutine]=None
    spinner_has_timer: bool=False
    spinner_enabled: bool=True
    spinner_display: ProgressText=None

    def from_dict(self, config: Dict[str, Any]):
        for config_value_name, config_value in config.items():
            if hasattr(self, config_value_name):
                setattr(self, config_value_name, config_value)

    @property
    def filesystem_logger(self):
        return {
            'logger_name': self.logger_name,
            'logger_type': self.logger_type,
            'logfiles_directory': self.logfiles_directory,
            'log_level': self.log_level,
            'logger_enabled': self.logger_enabled,
            'rotation_interval_type': self.filesystem_rotation_interval_type,
            'rotation_interval': self.filesystem_rotation_interval,
            'backup_count': self.filesystem_backup_count,
            'rotation_time': self.filesystem_rotation_time
        }

    @property
    def cli_logger(self):
        return {
            'logger_name': self.logger_name,
            'logger_type': self.logger_type,
            'log_level': self.log_level,
            'logger_enabled': self.logger_enabled,
        }

    @property
    def spinner(self):
        return {
            'logger_name': self.logger_name,
            'logger_type': self.logger_type,
            'log_level': self.log_level,
            'logger_enabled': self.logger_enabled,
            'spinner': self.spinner_type,
            'color': self.spinner_color,
            'on_color': self.spinner_on_color,
            'attrs': self.spinner_attrs,
            'reversal': self.spinner_reversal,
            'side': self.spinner_side,
            'sigmap': self.spinner_sigmap,
            'timer': self.spinner_has_timer,
            'enabled': self.spinner_enabled,
            'text': self.spinner_display
        }


from werkflow.modules.base import Module
from typing import Tuple, Type
from .validator import RequiresValidator


def requires(*modules: Tuple[Type[Module], ...]):

    RequiresValidator(
        modules=modules
    )

    def wraps(cls: Type[Module]):

        return cls
    
    return wraps


import distro

from werkflow.modules.system.architecture.types import ArchitectureType

from .base import OperatingSystem
from .types import OperatingSystemType


class Linux(OperatingSystem):

    def __init__(self) -> None:
        super().__init__()

        self.name = distro.id().capitalize()
        self.type = OperatingSystemType.LINUX

        self.supported_architectures = {
            ArchitectureType.ARM_64: 'arm64',
            ArchitectureType.x86_64: 'x86_64',
            ArchitectureType.AMD_64: 'AMD64',
            ArchitectureType.AARCH64: 'aarch64',
        }
        


import asyncio
import click
import functools
from werkflow.prompt.types.base.base_prompt import BasePrompt
from werkflow.prompt.types.base.prompt_type import PromptType
from typing import Callable, Union, Optional, Any, Dict
from .repeat_prompt_validator import RepeatPromptValidator


class RepeatPrompt(BasePrompt):

    def __init__(
        self,
        prompt: BasePrompt,
        result_key: str=None,
        skip: bool=False,
        condition: Optional[Callable[[Dict[str, Any]], bool]]=None,
        break_condition: Callable[[Dict[str, Any]], bool]=lambda response: response is not None,
        confirmation_message: Optional[Union[str, Callable[..., str]]] = None, 

    ) -> None:
        super().__init__(
            None,
            skip,
            result_key,
            None,
            condition,
            confirmation_message
        )

        validated_prompt = RepeatPromptValidator(
            prompt_repeated=prompt,
            prompt_break_condition=break_condition,
            prompt_type=PromptType.REPEAT
        )

        self.prompt: BasePrompt = validated_prompt.prompt_repeated
        self.break_condition = validated_prompt.prompt_break_condition  
        self.prompt_type = validated_prompt.prompt_type

    async def ask(self):
        
        responses = []
        response = None

        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        while not self.break_condition(response):

            if response is not None:
                responses.append(response)

            response = await self.prompt.ask()
            await self.prompt.confirm(response)

            if not self.break_condition(response):
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        click.echo,
                        ''
                    )
                )

        return responses

    async def confirm(self, response: Any):
        if self.confirmation_message:

            if self._loop is None:
                self._loop = asyncio.get_running_loop()
            
            if isinstance(self.confirmation_message, str):
                confirmation_text = self.confirmation_message

            else:
                confirmation_text = self.confirmation_message(response)

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    click.echo,
                    f'{self.prompt.confirmation_frame} {confirmation_text}',
                )
            )

    def close(self):
        self.prompt.close()


import click
import asyncio
import functools
from typing import List, Dict, Optional, Any, Callable, Union
from .option_prompt_validator import OptionPromptValidator
from werkflow.prompt.types.base.base_prompt import BasePrompt
from werkflow.prompt.types.base.prompt_type import PromptType


class OptionPrompt(BasePrompt):

    def __init__(
        self, 
        message: str, 
        skip: bool = False, 
        options: List[str] = None, 
        result_key: str = None, 
        default: Optional[Any] = None, 
        condition: Optional[Callable[[Dict[str, Any]], bool]] = None, 
        confirmation_message: Optional[Union[str, Callable[..., str]]] = None, 
    ) -> None:
        super().__init__(
            message, 
            skip, 
            result_key, 
            default, 
            condition, 
            confirmation_message,
        )

        validated_prompt = OptionPromptValidator(
            prompt_data_type=click.Choice(
                options,
                case_sensitive=False
            ),
            prompt_type=PromptType.OPTION
        )

        self.options = validated_prompt.prompt_data_type
        self.prompt_type = validated_prompt.prompt_type
        

    async def ask(self):
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

        return await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                click.prompt,
                f'{self.prompt_frame} {self.message}',
                type=self.options,
                default=self.default,
                prompt_suffix='',
                show_choices=True
            )
        )


import asyncio
import os
import signal
from werkflow.graph import Graph
from werkflow.logging import WerkflowLogger
from multiprocessing import current_process, active_children

def handle_loop_stop(
    signame,
    logger: WerkflowLogger,
    graph: Graph
):
        try:
            for workflow in graph._workflows.values():
                workflow.abort()
            child_processes = active_children()
            for child in child_processes:
                child.kill()
                
            process = current_process()
            if process:
                try:
                    process.kill()
                
                except Exception:
                    pass

        except BrokenPipeError:
            logger.console.sync.critical('\n\nAborted.\n')   

        except RuntimeError:
            logger.console.sync.critical('\n\nAborted.\n')

        if len(child_processes) < 1:
            logger.console.sync.critical('\n\nAborted.\n')   
            os._exit(1)

def add_abort_handler(
    loop: asyncio.AbstractEventLoop, 
    logger: WerkflowLogger,
    graph: Graph
):
    for signame in ('SIGINT', 'SIGTERM'):
        loop.add_signal_handler(
            getattr(signal, signame),
            lambda signame=signame: handle_loop_stop(
                signame, 
                logger,
                graph
            )
        )